#if !defined(__affinity_h__)
#define __affinity_h__ 1


//
// These affinity routines are simplified and only work with up to 64 CPUs.  They can be abstracted to
// more CPUs, but I currently dont care about machines that have more than 16 * 2 * 2 CPUS.
//
namespace aff
{
	class cpu_affinity_mask
	{
		const static int mycpumasksize = 64;
		uint64_t mask;
		cpu_set_t cpuset;
	public:
		cpu_affinity_mask&
		setbits(const uint64_t *srcarray, int nchunks)
		{
			for (int i = 1; i < nchunks; i++)
			{
				if (srcarray[i] != 0)
				{
					::fprintf(stderr,"Non-fatal warning: Source mask for cpu_affinity_mask::setbits\n");
					break;
				}
			}

			mask |= srcarray[0];
			return *this;
		}

		cpu_affinity_mask&
		clrbits(const uint64_t *srcarray, int nchunks)
		{
			for (int i = 1; i < nchunks; i++)
			{
				if (srcarray[i] != 0)
				{
					::fprintf(stderr,"Non-fatal warning: Source mask for cpu_affinity_mask::clrbits\n");
					break;
				}
			}

			mask &= ~srcarray[0];
			return *this;
		}

		cpu_affinity_mask&
		andbits(const uint64_t *srcarray, int nchunks)
		{
			for (int i = 1; i < nchunks; i++)
			{
				if (srcarray[i] != 0)
				{
					::fprintf(stderr,"Non-fatal warning: Source mask for cpu_affinity_mask::clrbits\n");
					break;
				}
			}

			mask &= srcarray[0];
			return *this;
		}



		cpu_affinity_mask&
		setbit(int bit)
		{
			if (bit >= (int) (8 * sizeof(mask)))
			{
				::fprintf(stderr,"Non-fatal warning: Source bit number for cpu_affinity_mask::setbit\n");
			}
			else
			{
				mask |= (1ULL << bit);
			}
			return *this;
		}

		cpu_affinity_mask&
		clrbit(int bit)
		{
			if (bit >= (int) (8 * sizeof(mask)))
			{
				::fprintf(stderr,"Non-fatal warning: Source bit number for cpu_affinity_mask::clrbit\n");
			}
			else
			{
				mask &= ~(1ULL << bit);
			}
			return *this;
		}

		cpu_affinity_mask&
		clrall(void)
		{
			mask = 0;
			return *this;
		}

		bool
		isbitset(int bit)
		{
			if (bit >= (int) (8 * sizeof(mask)))
			{
				::fprintf(stderr,"Non-fatal warning: Source bit number for cpu_affinity_mask::isbitset\n");
				return false;
			}
			else
			{
				return (mask & (1ULL << bit)) ? true : false;
			}
		}
			
		cpu_affinity_mask(void) : mask(~0), cpuset() {}

		cpu_affinity_mask(const cpu_affinity_mask& src) : mask(src.mask), cpuset() {}

		cpu_affinity_mask(const uint64_t* srcmask, int nchunks) : mask(0), cpuset()
		{
			setbits(srcmask, nchunks);
		}

		cpu_affinity_mask&
		operator=(const cpu_affinity_mask& src)
		{
			mask = src.mask;
			return *this;
		}

		cpu_affinity_mask&
		setmask(const uint64_t* src, int nchunks)
		{
			clrall();
			setbits(src, nchunks);
			return *this;
		}

		cpu_affinity_mask&
		addmask(const uint64_t *src, int nchunks)
		{
			setbits(src, nchunks);
			return *this;
		}

		cpu_affinity_mask&
		andmask(const uint64_t *src, int nchunks)
		{
			andbits(src, nchunks);
			return *this;
		}

		cpu_affinity_mask&
		clrmask(const uint64_t *src, int nchunks)
		{
			clrbits(src, nchunks);
			return *this;
		}

		cpu_set_t*
		cvt2cpu_set_t(void)
		{
			CPU_ZERO(&cpuset);
			for (int i = 0; i < mycpumasksize; i++)
			{
				if (isbitset(i))
					CPU_SET(i, &cpuset);
			}
			return &cpuset;
		}

		const cpu_set_t*
		cpu_set_tp(void) const
		{
			return &cpuset;
		}

		int
		size_cpu_set_t(void) const
		{
			return sizeof(cpuset);
		}

		cpu_affinity_mask&
		synchronize_set_to_mask(void)
		{
			memset(&mask, 0, sizeof mask);
			for (int i = 0; i < mycpumasksize; i++)
			{
				if (CPU_ISSET(i, &cpuset))
					mask |= (1ULL << i);
			}
			return *this;
		}

		int
		maxcpus(void) const
		{
			return mycpumasksize;
		}

		//
		// VERY BAD ROUTINE -- should not be used by anyone
		//
		uint64_t
		rawmask(void) const
		{
			return mask;
		}
	};


	class cpu_affinity
	{
		cpu_affinity_mask mask;

		const cpu_affinity&
		setstate(pid_t pid = ::getpid())
		{
			cpu_set_t *newset;
			newset = mask.cvt2cpu_set_t();
			int newsize;
			newsize = mask.size_cpu_set_t();

			int rtval = ::sched_setaffinity(pid, newsize, newset);
			if (rtval != 0)
			{
				::perror("sched_setaffinity");
			}
			return *this;
		}

		const cpu_affinity&
		getstate(pid_t pid = ::getpid())
		{
			cpu_set_t *newset;
			newset = mask.cvt2cpu_set_t();
			int newsize;
			newsize = mask.size_cpu_set_t();

			mask.clrall();

			int rtval = ::sched_getaffinity(pid, newsize, newset);
			if (rtval != 0)
			{
				::perror("sched_getaffinity");
			}
			else
			{
				mask.synchronize_set_to_mask();
			}
			return *this;
		}

		const cpu_affinity&
		pth_setstate(pthread_t thread = pthread_self())
		{
			cpu_set_t *newset;
			newset = mask.cvt2cpu_set_t();
			int newsize;
			newsize = mask.size_cpu_set_t();

			int rtval = ::pthread_setaffinity_np(thread, newsize, newset);
			if (rtval != 0)
			{
				::perror("pthread_setaffinity");
			}
			return *this;
		}

		const cpu_affinity&
		pth_getstate(pthread_t thread = pthread_self())
		{
			cpu_set_t *newset;
			newset = mask.cvt2cpu_set_t();
			int newsize;
			newsize = mask.size_cpu_set_t();

			mask.clrall();

			int rtval = ::pthread_getaffinity_np(thread, newsize, newset);
			if (rtval != 0)
			{
				::perror("pthread_getaffinity");
			}
			else
			{
				mask.synchronize_set_to_mask();
			}
			return *this;
		}


	public:
		cpu_affinity(void) NOINLINE : mask() {}

		cpu_affinity(const cpu_affinity& src) NOINLINE : mask(src.mask) {}

		const cpu_affinity& NOINLINE
		addcpu(int cpu)
		{
			mask.setbit(cpu);
			return *this;
		}

		const cpu_affinity& NOINLINE
		delcpu(int cpu)
		{
			mask.clrbit(cpu);
			return *this;
		}

		const NOINLINE
		cpu_set_t *getaffinity(pid_t pid = ::getpid())
		{
			getstate(pid);
			return mask.cpu_set_tp();
		}

		//
		// This interface is slightly defective - only works with up to 64 CPUS,
		// but can easily be upgraded.
		//
		const NOINLINE
		cpu_set_t *setaffinity(uint64_t maskarg, pid_t pid = ::getpid())
		{
			mask.setmask(&maskarg, 1);
			setstate(pid);
			return getaffinity(pid);
		}

		const cpu_set_t * NOINLINE
		pth_getaffinity(pthread_t thread = pthread_self())
		{
			pth_getstate(thread);
			return mask.cpu_set_tp();
		}
			
		const cpu_affinity& NOINLINE
		pth_setaffinity(uint64_t maskarg, pthread_t thread = pthread_self())
		{
			mask.setmask(&maskarg, 1);
			pth_setstate(thread);
			return pth_getstate(thread);
		}

		void NOINLINE
		pth_printaffinity(FILE *fp, int index, pthread_t thread = pthread_self())
		{
			const cpu_set_t* UNUSED aff = pth_getaffinity(thread);

			::fprintf(fp, "affinitymask(tindex: %d, tid: %p): %8.8llx\n", index, (void *) thread, (unsigned long long) mask.rawmask());
		}

	};
};

#endif

//
// vim: set ts=4:
//

