//
// provide low level interface definitions.  Most of the real-IPC interface
// is using JSON -- which is also partially defined here.
//


namespace agc
{
	static constexpr int AGCUNAMESIZE = 16;
	static constexpr int POSTNAMESIZE = 4;
	static constexpr int NCLICKSAMPLES = 4;
	static constexpr int NPREVIOUSSAMPLES = 10;
	static constexpr int nclicksinusec[NCLICKSAMPLES] =
	{
		1000,		// 1msec
		10000,		// 10msec
		100000,		// 100msec
		1000000		// 1000msec
	};

	struct gainstats
	{
		//
		// we remember the last 'NPREVIOUSSAMPLES' of each
		// of the 'NCLICKSAMPLES' sample rate averages.
		// this can allow for better stats without taking
		// huge amounts of CPU.
		//
		int cursample[NCLICKSAMPLES];
		float curgainLIN[NPREVIOUSSAMPLES][NCLICKSAMPLES];
	public:
		gainstats(void)
		{
			for (int j = 0; j < NPREVIOUSSAMPLES; j++)
			{
				for (int i = 0; i < NCLICKSAMPLES; i++)
				{
					curgainLIN[j][i] = 0.0f;
				}
			}
		}
	};

	struct levelstats
	{
		//
		// we remember the last 'NPREVIOUSSAMPLES' of each
		// of the 'NCLICKSAMPLES' sample rate averages.
		// this can allow for better stats without taking
		// huge amounts of CPU.
		//
		int cursample[NCLICKSAMPLES];
		float curlevelLIN[NPREVIOUSSAMPLES][NCLICKSAMPLES];
		levelstats(void)
		{
			for (int j = 0; j < NPREVIOUSSAMPLES; j++)
			{
				for (int i = 0; i < NCLICKSAMPLES; i++)
				{
					curlevelLIN[j][i] = 0.0f;
				}
			}
		}
	};

	struct unitstatus
	{
		int nclickstates[NCLICKSAMPLES];
		float nclickfilters[NCLICKSAMPLES];
		gainstats gain;
		levelstats level;
	};

		
		
	struct agcunit
	{
		char unitname[AGCUNAMESIZE];
		char alwayszero0[POSTNAMESIZE];
		int unitnameunique;
		char unitnickname[AGCUNAMESIZE];
		char alwayszero1[POSTNAMESIZE];
		float inpgaindB;	// gain before unit
		float maxgaindB;	// max gain by unit
		float crdBperdB;	// compression ratio
		float thresholddB;	// threshold re: full scale
		float gatedB;		// gate where decay slows down
		float bluntdepthdB;	// input range where attack is ignored
		float bluntpower;	// exponential power of fast attacks
					// (smaller means less effect 1.0 max)

	public:
		agcunit(void) :
			unitname(), alwayszero0(), unitnameunique(0),
			unitnickname(), alwayszero1(),
			inpgaindB(0.0f), maxgaindB(0.0f), crdBperdB(1.0f),
			thresholddB(0.0f), gatedB(0.0f),
			bluntdepthdB(0.0f), bluntpower(1.0f)

		agcunit(const char *un, int ununique,
			const char *uun, int unnkunique,
			float igdB = 0.0f, float mgdB = 0.0f, float cr = 1.0f,
			float threshdB = 0.0f, float gdB = 0.0f, float bddB = 0.0f, float bpower = 0.10f):
			unitname(), alwayszero0(), unitnameunique(ununique),
			unitnickname(), alwayszero1(),
			inputgaindB(igdB), maxgaindB(mgdB), crdBperdB(cr),
			thresholddB(threshdB), bluntdepthdB(bddB), bluntpower(bpower)
		{
			strncpy(unitname, un, AGCNAMESIZE - 1);
			strncpy(unitnickname, uun, AGCNAMESIZE - 1);
		}
	
	};

	//
	// So-called 'release' times are set by the speed paramters of each phase of the
	// AGC.  The simple idea of 'release time' or especially 'attack time' are meaningless
	// on this system.  It is quite capable of following the envelope of the
	// audio waveform with minimal intermod (or harmonic distortion.)  Initially
	// I tried some hilbert techniques (a step beyond RMS), but found a better
	// method yet.
	//
	// The decay rates are based upon the desire of certain kinds of sound quality.
	// There are almost no cases where one might specify a rate like '1dB/sec' or
	// somesuch.  There is one parameter that can change some of the preprocessors
	// rates to be as slow as '1dB/sec' or '4dB/sec', but those numbers are relatively
	// meaningless -- because the attack/decay rates are typically going to range
	// anywhere between just above 1000dB/sec to 1dB/sec or slower.  There are also
	// some nonlinear determinations which can stop the release which effectively
	// stop the gain increase at certain times.
	//
	// However, there is a significant amount of speed tuning that can be done, it just
	// doesn't sound like other AGC systems.  Even using the manual parameters, it is
	// difficult to get a severe 'squeezed' sound effect.  If requested, I can add
	// a mode that could produce a variable amount of 'squeezed' sound.
	//
	// This processor uses both linear and dB oriented AGC attacks/decays, where the majority
	// of level control is in the dB domain, but certain phases are predominantly old-fashioned
	// linear compression.  One problem with normal linear domain compression that is that
	// it is too easy to get a 'squeezed' sound.  So, the largest amounts of gain control
	// are done in the dB domain, but short attacks can be overridden by the linear domain.
	// Some of the very fast AGC (with just a few dB max of control)
	// is done in the linear domain.)  The fastest attack/decay rates are effectively zero
	// for the attack time (partially synchronized with the waveform), and the decay time
	// is the same as the attack slew of the signal envelope.  (That is, the attack/decay
	// times can be very fast -- but one only sees the super fast behavior with short impulses.)
	//
	// One of the goals is to mitigate the intermod effects and other horrible side effects
	// so often heard on older recordings (and some new recordings.)  The obvious way to help
	// is to carefully control levels before processing.  That fact is still true, but
	// this processor really does the minimum damage that I have been able to figure out
	// given the amount of available processing.  It might be possible to do more with less
	// disruption in the frequency domain and I am researching that.
	//

	//
	// Note that the units below are not as redudant as they might seem.
	// For example, the PREAGC1 input is preprocessed by PREAGC0 to make sure that
	// the multi band AGC doesn't cause too many side-effects.
	//
	// AGC0, AGC1 are similarly benefitted by PREAGC2.  Additionally, AGC0, AGC1
	// are benefitted by all previous stages.  Too much multi-band AGC can flatten
	// the spectrum TOO MUCH.  I worked for quite a while to correct the effects
	// of the severe spectrum flattening, and simply learned that multi-band AGC
	// needs to be reigned-in.
	//
	// QPRELIMIT uses an algorithm that both causes minimal disturbance and does a good
	// job of minimizing the work that the PRELIMIT needs to do.  THe goal of QPRELIMIT
	// and QLIMIT are to help to avoid intermod, yet achieve the goal of FAST gain control.
	//
	// At the point of the QLIMIT input, the signal has most of its final character, other
	// than safely the limiter chain needs to deal with peaks/transients/etc.
	// The QLIMIT does the limiting job on a spectrum basis, and makes sure that
	// severe transients are mitigated without causing glitching.
	//
	// At the point of the limiter, it can do most of the hard limiting with only at most
	// usually less than 1.5dB of gain control -- mostly on the hardest peaks.  however,
	// such severe agc still leaves some peaks, and also the normal 'LIMITER' is set at
	// fairly low compression ratio (abt 8:1) to maintain the best possible sound quality.
	//
	// The final hard limiter is usually hit with a fraction of a dB -- finally cleaning
	// up most of the mess from the limiter.  If the signal passed through to the limiter
	// instead of using this final hard limiter, the sound quality really suffers.  Of course,
	// the normal last limiter COULD do the job of both the LIMITER and hard limiter, but
	// it is difficult to control the last bit of limiting, so this one mostly does
	// super quick things.  (Can handle LF if needed, however.)
	//
	// Then, the output threshold determines the output clipping.  Unfortunately, the clipping
	// does happen when hitting the system hard -- but it is entirely un-noticable because
	// the final hard limiter does the real cleanup (except for a few samples or a few 0.01dB.)
	//
	// The only subsystems that have significant independent controls are:
	//
	// INPUT, PREAGC0, PREAGC1, PREAGC2, AGC1, AGC2, PRELIMIT, LIMITER and OUTPUT.
	//
	// FOR tunable COMPRESSION RATIOS:
	//
	// Usually, PREAGC0, PREAGC2 have very limited CR of about 1.0-1.7
	// PREAGC1 CR 1.0-2.0
	// AGC0 CR 1.0-3.0
	// AGC1 CR 1.0-1.7
	//
	// PRELIMIT CR 3-5
	// LIMITER CR 5-100
	// OUTPUT CLIPPING ONLY
	//
	// QPRELIMIT is connected to PRELIMITs parameters and takes a large amount of the PRELIMITER
	// burden when enabled.
	//
	// QLIMIT is connected with the LIMITERs parameters, and also takes a large about of the LIMITER
	// burden.
	//
	// HLIMIT has inf:1 compression ratio -- will all but clip when needed.
	//
	// OUTPUT mostly clips due to gibbs effect.
	//

	//
	// Configuration differences:
	//
	// Usually, the only subsystem which benefits from maxgain > 0.0dB is the AGC1, which does a good
	// job of filling in between transients.  Using maxgain > 0 on other subsystems usually just
	// increases pumping or background noise.
	//
	// The gain parameter for each subsystem is for gain BEFORE that subsystem.
	// The subsystems that mostly just have measurements are:  QPRELIMIT, QLIMIT and HLIMIT.
	// 
	// INPUT only has a gain parameter.
	//
	// OUTPUT has a gain and threshold parameter (clipping level.)
	//
	// HLIMIT, QLIMIT and QPRELIMIT derive their parameters from other subsystems.
	//  (There is little/no benefit of tuning these subsystems other than enable/disable)
	// 
	// HLIMIT, QLIMIT and QPRELIMIT can be individually disabled.
	//
	// PREAGC0, PREAGC1, PREAGC2, AGC0, AGC1, PRELIMIT, LIMITER are disabled by setting
	//  their compression ratios (CR) to 1:1.
	//
	//
	// Many of the subsystems run in different threads (some are dependent on multiple threads so as
	// to maintain maximum concurrency.)  So, some measurements will seem a little out of sync.
	//
	// The UI will provide a coordinated control so that many individual controls will be optionally
	// synchronized into a few simplified controls.
	//
	// The processor supports 3 sample rates: 44100, 48000, 96000.
	// At 96000, a quad Haswell CPU will utilize about 1/2 of the total CPU power -- but
	// at 44.1k or 48k, it will tend to utilize the power of approx 1 core.
	// (I have some optimizations in mind -- which will not materially affect the AGC algorithms,
	// but might change the character of the multi-band filtering.)  Currently, the processor does not
	// use subsampling, but would probably benefit to using about 1/3-1/2 of the CPU if subsampling
	// was used. The audible effects would have be be carefully evaluated, however.
	//
	// The processor does sound a little different between 44.1k, 48k and 96k, where 96k does sound
	// a little bit better.
	//
	// The processor supports .wav file I/O with 32bit FP, or 16bit signed integer at any of the 3
	// baud rates.
	//
	// If the CPU is fast enough, can run in realtime, but CURRENTLY does have a significant delay
	// due to multiple levels of multi-band filtering, multi-thread buffering, and some minor explicit
	// delays to synchronize the audio and control channels.
	//

	enum class AGCUNITIDX = {
		INPUT = 0,	// NULL UNIT -- just gain
		PREAGC0 = 1,	// Handles signal level range compatibility (LINEAR + dB)
		PREAGC1 = 2,	// Three band -- handles early AGC (LINEAR + dB)
		PREAGC2 = 3,	// Helps to stabilize the output of the three band (LINEAR + dB)
		AGC0 = 4,	// Main multiband -- dB oriented (LINEAR + dB)
		AGC1 = 5,	// Secondary (fast) multiband -- LINEAR ONLY
		QPRELIMIT = 6	// quick prelimit -- does backwards multiband prelimiting
		PRELIMIT = 7,	// Pre-limiter -- stabilizes output of multiband (LINEAR + dB)
		QLIMIT = 8,	// quick limit -- does backwards multiband limiting
		LIMITER = 9,	// Real limiter -- does most of the hard limiting
		HLIMIT = 10,	// Super hard/last chance limiter
		OUTPUT = 11,	// Output gain/clipper
		NUNITS = 12
	};

	static agcunit unitinfo[AGCUNITIDX::NUNITS] = {
		{ "input", 3, "inp"},		// input gain
		{ "preagc0", 6, "pa0"},		// agc before 3 band
		{ "preagc1", 6, "pa1"},		// 3band AGC
		{ "preagc2", 6, "pa2"},		// agc after 3 band AGC
		{ "agc0", 4, "a0"},		// multi (8) band compression,
		{ "agc1", 4, "a1"},		// fast multiband compression
		{ "qprelimit", 4, "qpl"},	// quick prelimit (special backwards algorithm)
		{ "prelimit", 4, "pl"},		// true prelimit
		{ "qlimit", 3, "ql"}		// quick limit (special backwards algorithm)
		{ "limit", 3, "lm"},		// limit
		{ "hlimit", 3, hl"},		// last chance limit
		{ "output", 3, "out"}		// output gain/clipping
	};

};

